import os
from gi.repository import GtkSource

from Library import Library

global manager
manager = None

def get_language_manager():
        global manager
        
        if not manager:
                dirs = []
        
                for d in Library().systemdirs:
                        dirs.append(os.path.join(d, 'lang'))
        
                manager = GtkSource.LanguageManager()
                manager.set_search_path(dirs + manager.get_search_path())
        
        return manager



# -*- coding: utf-8 -*-

#  Copyright (C) 2009 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from gi.repository import Gio, Gtk

class VirtualDirectory(object):
        def __init__(self, name):
                self._name = name
                self._children = []

        def get_uri(self):
                return 'virtual://' + self._name

        def get_parent(self):
                return None

        def enumerate_children(self, attr, flags, callback):
                return self._children

        def append(self, child):
                if not child.is_native():
                        return

                try:
                        info = child.query_info("standard::*", Gio.FileQueryInfoFlags.NONE, None)

                        if info:
                                self._children.append((child, info))
                except:
                        pass

class RecentDocumentsDirectory(VirtualDirectory):
        def __init__(self, maxitems=10):
                VirtualDirectory.__init__(self, 'recent')

                self._maxitems = maxitems
                self.fill()

        def fill(self):
                manager = Gtk.RecentManager.get_default()

                items = manager.get_items()
                items.sort(lambda a, b: cmp(b.get_visited(), a.get_visited()))

                added = 0

                for item in items:
                        if item.has_group('pluma'):
                                self.append(Gio.file_new_for_uri(item.get_uri()))
                                added += 1

                                if added >= self._maxitems:
                                        break

class CurrentDocumentsDirectory(VirtualDirectory):
        def __init__(self, window):
                VirtualDirectory.__init__(self, 'documents')

                self.fill(window)

        def fill(self, window):
                for doc in window.get_documents():
                        location = doc.get_location()
                        if location:
                                self.append(location)

# ex:ts=8:et:



# -*- coding: utf-8 -*-

#  Copyright (C) 2009 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from gi.repository import GObject, Peas
from windowhelper import WindowHelper

class QuickOpenPlugin(GObject.Object, Peas.Activatable):
        __gtype_name__ = "QuickOpenPlugin"

        object = GObject.Property(type=GObject.Object)

        def __init__(self):
                GObject.Object.__init__(self)

                self._popup_size = (450, 300)

        def do_activate(self):
                window = self.object
                self._helper = WindowHelper(window, self)

        def do_deactivate(self):
                self._helper.deactivate()
                self._helper = None

        def do_update_state(self):
                self._helper.update_ui()

        def get_popup_size(self):
                return self._popup_size

        def set_popup_size(self, size):
                self._popup_size = size

# ex:ts=8:et:



# -*- coding: utf-8 -*-

#  Copyright (C) 2009 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import os
import fnmatch
import xml.sax.saxutils
from gi.repository import GObject, Gio, GLib, Gdk, Gtk, Pango, Pluma
from virtualdirs import VirtualDirectory

class Popup(Gtk.Dialog):
        __gtype_name__ = "QuickOpenPopup"

        def __init__(self, window, paths, handler):
                Gtk.Dialog.__init__(self,
                                    title=_('Quick Open'),
                                    parent=window,
                                    flags=Gtk.DialogFlags.DESTROY_WITH_PARENT | Gtk.DialogFlags.MODAL,
                                    buttons=(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL))

                self._open_button = self.add_button(Gtk.STOCK_OPEN, Gtk.ResponseType.ACCEPT)

                self._handler = handler
                self._build_ui()

                self._size = (0, 0)
                self._dirs = []
                self._cache = {}
                self._theme = None
                self._cursor = None
                self._shift_start = None

                self._busy_cursor = Gdk.Cursor(Gdk.CursorType.WATCH)

                accel_group = Gtk.AccelGroup()
                accel_group.connect(Gdk.KEY_l, Gdk.ModifierType.CONTROL_MASK, 0, self.on_focus_entry)

                self.add_accel_group(accel_group)

                unique = []

                for path in paths:
                        if not path.get_uri() in unique:
                                self._dirs.append(path)
                                unique.append(path.get_uri())

        def get_final_size(self):
                return self._size

        def _build_ui(self):
                vbox = self.get_content_area()
                vbox.set_spacing(3)

                self._entry = Gtk.Entry()

                self._entry.connect('changed', self.on_changed)
                self._entry.connect('key-press-event', self.on_key_press_event)

                sw = Gtk.ScrolledWindow()
                sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
                sw.set_shadow_type(Gtk.ShadowType.OUT)

                tv = Gtk.TreeView()
                tv.set_headers_visible(False)

                self._store = Gtk.ListStore(Gio.Icon, str, GObject.Object, Gio.FileType)
                tv.set_model(self._store)

                self._treeview = tv
                tv.connect('row-activated', self.on_row_activated)

                renderer = Gtk.CellRendererPixbuf()
                column = Gtk.TreeViewColumn()
                column.pack_start(renderer, False)
                column.add_attribute(renderer, "gicon", 0)

                renderer = Gtk.CellRendererText()
                column.pack_start(renderer, True)
                column.add_attribute(renderer, "markup", 1)

                column.set_cell_data_func(renderer, self.on_cell_data_cb, None)

                tv.append_column(column)
                sw.add(tv)
                
                selection = tv.get_selection()
                selection.connect('changed', self.on_selection_changed)
                selection.set_mode(Gtk.SelectionMode.MULTIPLE)

                vbox.pack_start(self._entry, False, False, 0)
                vbox.pack_start(sw, True, True, 0)

                lbl = Gtk.Label()
                lbl.set_alignment(0, 0.5)
                lbl.set_ellipsize(Pango.EllipsizeMode.MIDDLE)
                self._info_label = lbl

                vbox.pack_start(lbl, False, False, 0)

                # Initial selection
                self.on_selection_changed(tv.get_selection())
                vbox.show_all()

        def on_cell_data_cb(self, column, cell, model, piter, user_data):
                path = model.get_path(piter)
                
                if self._cursor and path == self._cursor.get_path():
                        style = self._treeview.get_style()
                        bg = style.bg[Gtk.StateType.PRELIGHT]
                        
                        cell.set_property('cell-background-gdk', bg)
                        cell.set_property('style', Pango.Style.ITALIC)
                else:
                        cell.set_property('cell-background-set', False)
                        cell.set_property('style-set', False)

        def _icon_from_stock(self, stock):
                theme = Gtk.icon_theme_get_default()
                size = Gtk.icon_size_lookup(Gtk.IconSize.MENU)
                pixbuf = theme.load_icon(stock, size[0], Gtk.IconLookupFlags.USE_BUILTIN)

                return pixbuf

        def _list_dir(self, gfile):
                entries = []

                try:
                        ret = gfile.enumerate_children("standard::*", Gio.FileQueryInfoFlags.NONE, None)
                except GLib.GError:
                        pass

                if isinstance(ret, Gio.FileEnumerator):
                        while True:
                                entry = ret.next_file(None)

                                if not entry:
                                        break

                                entries.append((gfile.get_child(entry.get_name()), entry))
                else:
                        entries = ret

                children = []

                for entry in entries:
                        children.append((entry[0], entry[1].get_name(), entry[1].get_file_type(), entry[1].get_icon()))

                return children

        def _compare_entries(self, a, b, lpart):
                if lpart in a:
                        if lpart in b:
                                return cmp(a.index(lpart), b.index(lpart))
                        else:
                                return -1
                elif lpart in b:
                        return 1
                else:
                        return 0

        def _match_glob(self, s, glob):
                if glob:
                        glob += '*'

                return fnmatch.fnmatch(s, glob)

        def do_search_dir(self, parts, d):
                if not parts or not d:
                        return []

                if not d in self._cache:
                        entries = self._list_dir(d)
                        entries.sort(lambda x, y: cmp(x[1].lower(), y[1].lower()))

                        self._cache[d] = entries
                else:
                        entries = self._cache[d]

                found = []
                newdirs = []

                lpart = parts[0].lower()

                for entry in entries:
                        if not entry:
                                continue

                        lentry = entry[1].lower()

                        if not lpart or lpart in lentry or self._match_glob(lentry, lpart):
                                if entry[2] == Gio.FileType.DIRECTORY:
                                        if len(parts) > 1:
                                                newdirs.append(entry[0])
                                        else:
                                                found.append(entry)
                                elif entry[2] == Gio.FileType.REGULAR and \
                                     (not lpart or len(parts) == 1):
                                        found.append(entry)

                found.sort(lambda a, b: self._compare_entries(a[1].lower(), b[1].lower(), lpart))

                if lpart == '..':
                        newdirs.append(d.get_parent())

                for dd in newdirs:
                        found.extend(self.do_search_dir(parts[1:], dd))

                return found

        def _replace_insensitive(self, s, find, rep):
                out = ''
                l = s.lower()
                find = find.lower()
                last = 0

                if len(find) == 0:
                        return xml.sax.saxutils.escape(s)

                while True:
                        m = l.find(find, last)

                        if m == -1:
                                break
                        else:
                                out += xml.sax.saxutils.escape(s[last:m]) + rep % (xml.sax.saxutils.escape(s[m:m + len(find)]),)
                                last = m + len(find)

                return out + xml.sax.saxutils.escape(s[last:])


        def make_markup(self, parts, path):
                out = []

                for i in range(0, len(parts)):
                        out.append(self._replace_insensitive(path[i], parts[i], "<b>%s</b>"))

                return os.sep.join(out)

        def _get_icon(self, f):
                query = f.query_info(Gio.FILE_ATTRIBUTE_STANDARD_ICON, Gio.FileQueryInfoFlags.NONE, None)

                if not query:
                        return None
                else:
                        return query.get_icon()

        def _make_parts(self, parent, child, pp):
                parts = []

                # We went from parent, to child, using pp
                idx = len(pp) - 1

                while idx >= 0:
                        if pp[idx] == '..':
                                parts.insert(0, '..')
                        else:
                                parts.insert(0, child.get_basename())
                                child = child.get_parent()

                        idx -= 1

                return parts

        def normalize_relative(self, parts):
                if not parts:
                        return []

                out = self.normalize_relative(parts[:-1])

                if parts[-1] == '..':
                        if not out or (out[-1] == '..') or len(out) == 1:
                                out.append('..')
                        else:
                                del out[-1]
                else:
                        out.append(parts[-1])

                return out

        def _append_to_store(self, item):
                if not item in self._stored_items:
                        self._store.append(item)
                        self._stored_items[item] = True

        def _clear_store(self):
                self._store.clear()
                self._stored_items = {}

        def _show_virtuals(self):
                for d in self._dirs:
                        if isinstance(d, VirtualDirectory):
                                for entry in d.enumerate_children("standard::*", 0, None):
                                        self._append_to_store((entry[1].get_icon(), xml.sax.saxutils.escape(entry[1].get_name()), entry[0], entry[1].get_file_type()))

        def _set_busy(self, busy):
                if busy:
                        self.get_window().set_cursor(self._busy_cursor)
                else:
                        self.get_window().set_cursor(None)

                Gdk.flush()

        def _remove_cursor(self):
                if self._cursor:
                        path = self._cursor.get_path()
                        self._cursor = None

                        self._store.row_changed(path, self._store.get_iter(path))

        def do_search(self):
                self._set_busy(True)
                self._remove_cursor()

                text = self._entry.get_text().strip()
                self._clear_store()

                if text == '':
                        self._show_virtuals()
                else:
                        parts = self.normalize_relative(text.split(os.sep))
                        files = []

                        for d in self._dirs:
                                for entry in self.do_search_dir(parts, d):
                                        pathparts = self._make_parts(d, entry[0], parts)
                                        self._append_to_store((entry[3], self.make_markup(parts, pathparts), entry[0], entry[2]))

                piter = self._store.get_iter_first()

                if piter:
                        self._treeview.get_selection().select_path(self._store.get_path(piter))

                self.get_window().set_cursor(None)
                self._set_busy(False)

        def do_show(self):
                Gtk.Window.do_show(self)

                self._entry.grab_focus()
                self._entry.set_text("")

                self.do_search()

        def on_changed(self, editable):
                self.do_search()
                self.on_selection_changed(self._treeview.get_selection())

        def _shift_extend(self, towhere):
                selection = self._treeview.get_selection()
                
                if not self._shift_start:
                        model, rows = selection.get_selected_rows()
                        start = rows[0]

                        self._shift_start = Gtk.TreeRowReference(self._store, start)
                else:
                        start = self._shift_start.get_path()

                selection.unselect_all()
                selection.select_range(start, towhere)

        def _select_index(self, idx, hasctrl, hasshift):
                path = (idx,)
                
                if not (hasctrl or hasshift):
                        self._treeview.get_selection().unselect_all()
                
                if hasshift:
                        self._shift_extend(path)
                else:
                        self._shift_start = None
                        
                        if not hasctrl:
                                self._treeview.get_selection().select_path(path)

                self._treeview.scroll_to_cell(path, None, True, 0.5, 0)
                self._remove_cursor()

                if hasctrl or hasshift:
                        self._cursor = Gtk.TreeRowReference(self._store, path)
                        
                        piter = self._store.get_iter(path)
                        self._store.row_changed(path, piter)

        def _move_selection(self, howmany, hasctrl, hasshift):
                num = self._store.iter_n_children(None)

                if num == 0:
                        return True

                # Test for cursor
                path = None
                
                if self._cursor:
                        path = self._cursor.get_path()
                else:
                        model, rows = self._treeview.get_selection().get_selected_rows()
                        
                        if len(rows) == 1:
                                path = rows[0]

                if not path:
                        if howmany > 0:
                                self._select_index(0, hasctrl, hasshift)
                        else:
                                self._select_index(num - 1, hasctrl, hasshift)
                else:
                        idx = path[0]

                        if idx + howmany < 0:
                                self._select_index(0, hasctrl, hasshift)
                        elif idx + howmany >= num:
                                self._select_index(num - 1, hasctrl, hasshift)
                        else:
                                self._select_index(idx + howmany, hasctrl, hasshift)

                return True

        def _direct_file(self):
                uri = self._entry.get_text()
                gfile = None

                if Pluma.utils_is_valid_uri(uri):
                        gfile = Gio.file_new_for_uri(uri)
                elif os.path.isabs(uri):
                        f = Gio.file_new_for_uri(uri)

                        if f.query_exists():
                                gfile = f

                return gfile

        def _activate(self):
                model, rows = self._treeview.get_selection().get_selected_rows()
                ret = True
                
                for row in rows:
                        s = model.get_iter(row)
                        info = model.get(s, 2, 3)

                        if info[1] != Gio.FileType.DIRECTORY:
                                ret = ret and self._handler(info[0])
                        else:
                                text = self._entry.get_text()

                                for i in range(len(text) - 1, -1, -1):
                                        if text[i] == os.sep:
                                                break

                                self._entry.set_text(os.path.join(text[:i], os.path.basename(info[0].get_uri())) + os.sep)
                                self._entry.set_position(-1)
                                self._entry.grab_focus()
                                return True

                if rows and ret:
                        self.destroy()

                if not rows:
                        gfile = self._direct_file()

                        if gfile and self._handler(gfile):
                                self.destroy()
                        else:
                                ret = False
                else:
                        ret = False

                return ret

        def toggle_cursor(self):
                if not self._cursor:
                        return
                
                path = self._cursor.get_path()
                selection = self._treeview.get_selection()
                
                if selection.path_is_selected(path):
                        selection.unselect_path(path)
                else:
                        selection.select_path(path)

        def on_key_press_event(self, widget, event):
                move_mapping = {
                        Gdk.KEY_Down: 1,
                        Gdk.KEY_Up: -1,
                        Gdk.KEY_Page_Down: 5,
                        Gdk.KEY_Page_Up: -5
                }
                
                if event.keyval == Gdk.KEY_Escape:
                        self.destroy()
                        return True
                elif event.keyval in move_mapping:
                        return self._move_selection(move_mapping[event.keyval], event.state & Gdk.ModifierType.CONTROL_MASK, event.state & Gdk.ModifierType.SHIFT_MASK)
                elif event.keyval in [Gdk.KEY_Return, Gdk.KEY_KP_Enter, Gdk.KEY_Tab, Gdk.KEY_ISO_Left_Tab]:
                        return self._activate()
                elif event.keyval == Gdk.KEY_space and event.state & Gdk.ModifierType.CONTROL_MASK:
                        self.toggle_cursor()

                return False

        def on_row_activated(self, view, path, column):
                self._activate()

        def do_response(self, response):
                if response != Gtk.ResponseType.ACCEPT or not self._activate():
                        self.destroy()

        def do_configure_event(self, event):
                if self.get_realized():
                        alloc = self.get_allocation()
                        self._size = (alloc.width, alloc.height)

                return Gtk.Dialog.do_configure_event(self, event)

        def on_selection_changed(self, selection):
                model, rows = selection.get_selected_rows()
                
                gfile = None
                fname = None

                if not rows:
                        gfile = self._direct_file()
                elif len(rows) == 1:
                        gfile = model.get(model.get_iter(rows[0]), 2)[0]
                else:
                        fname = ''

                if gfile:
                        if gfile.is_native():
                                fname = xml.sax.saxutils.escape(gfile.get_path())
                        else:
                                fname = xml.sax.saxutils.escape(gfile.get_uri())

                self._open_button.set_sensitive(fname != None)
                self._info_label.set_markup(fname or '')

        def on_focus_entry(self, group, accel, keyval, modifier):
                self._entry.grab_focus()

# ex:ts=8:et:



# -*- coding: utf-8 -*-

# __init__.py -- plugin object
#
# Copyright (C) 2006 - Steve Frécinaux
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

# Parts from "Interactive Python-GTK Console" (stolen from epiphany's console.py)
#     Copyright (C), 1998 James Henstridge <james@daa.com.au>
#     Copyright (C), 2005 Adam Hooper <adamh@densi.com>
# Bits from pluma Python Console Plugin
#     Copyrignt (C), 2005 Raphaël Slinckx

from gi.repository import GObject, Gtk, Peas, Pluma

from console import PythonConsole
from config import PythonConsoleConfigDialog
from config import PythonConsoleConfig

PYTHON_ICON = 'text-x-python'

class PythonConsolePlugin(GObject.Object, Peas.Activatable):
    __gtype_name__ = "PythonConsolePlugin"

    object = GObject.Property(type=GObject.Object)

    def __init__(self):
        GObject.Object.__init__(self)
        self.dlg = None

    def do_activate(self):
        window = self.object

        self._console = PythonConsole(namespace = {'__builtins__' : __builtins__,
                                             'pluma' : Pluma,
                                             'window' : window})
        self._console.eval('print "You can access the main window through ' \
                     '\'window\' :\\n%s" % window', False)
        bottom = window.get_bottom_panel()
        image = Gtk.Image()
        image.set_from_icon_name(PYTHON_ICON, Gtk.IconSize.MENU)
        bottom.add_item(self._console, _('Python Console'), image)

    def do_deactivate(self):
        window = self.object

        self._console.stop()
        bottom = window.get_bottom_panel()
        bottom.remove_item(self._console)

def create_configure_dialog(self):

    if not self.dlg:
        self.dlg = PythonConsoleConfigDialog(self.get_data_dir())

    dialog = self.dlg.dialog()
    window = pluma.app_get_default().get_active_window()
    if window:
        dialog.set_transient_for(window)

    return dialog

# Here we dynamically insert create_configure_dialog based on if configuration
# is enabled. This has to be done like this because pluma checks if a plugin
# is configurable solely on the fact that it has this member defined or not
if PythonConsoleConfig.enabled():
    PythonConsolePlugin.create_configure_dialog = create_configure_dialog

# ex:et:ts=4:



#    Pluma snippets plugin
#    Copyright (C) 2005-2006  Jesse van den Kieboom <jesse@icecrew.nl>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
from gi.repository import GObject, GLib, Gtk, Peas, Pluma

from WindowHelper import WindowHelper
from Library import Library
from Manager import Manager

class SnippetsPlugin(GObject.Object, Peas.Activatable):
        __gtype_name__ = "SnippetsPlugin"

        object = GObject.Property(type=GObject.Object)

        def __init__(self):
                GObject.Object.__init__(self)

                self.dlg = None

        def system_dirs(self):
                if 'XDG_DATA_DIRS' in os.environ:
                        datadirs = os.environ['XDG_DATA_DIRS']
                else:
                        datadirs = '/usr/local/share' + os.pathsep + '/usr/share'

                dirs = []

                for d in datadirs.split(os.pathsep):
                        d = os.path.join(d, 'pluma', 'plugins', 'snippets')

                        if os.path.isdir(d):
                                dirs.append(d)

                dirs.append(self.plugin_info.get_data_dir())
                return dirs

        def do_activate(self):
                library = Library()
                library.add_accelerator_callback(self.accelerator_activated)

                snippetsdir = os.path.join(GLib.get_user_config_dir(), '/pluma/snippets')
                library.set_dirs(snippetsdir, self.system_dirs())

                self._helper = WindowHelper(self)

                window = self.object
                self._helper.run(window)

        def do_deactivate(self):
                library = Library()
                library.remove_accelerator_callback(self.accelerator_activated)

                self._helper.stop()
                self._helper = None

        def do_update_state(self):
                self._helper.update()

        def create_configure_dialog(self):
                if not self.dlg:
                        self.dlg = Manager(self.plugin_info.get_data_dir())
                else:
                        self.dlg.run()

                window = Pluma.App.get_default().get_active_window()

                if window:
                        self.dlg.dlg.set_transient_for(window)

                return self.dlg.dlg

        def accelerator_activated(self, group, obj, keyval, mod):
                ret = False

                if _self.helper:
                        ret = self._helper.accelerator_activated(keyval, mod)

                return ret



# -*- coding: utf-8 -*-

#  Copyright (C) 2009 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from popup import Popup
import os
from gi.repository import Gio, GLib, Gtk, Pluma
from virtualdirs import RecentDocumentsDirectory
from virtualdirs import CurrentDocumentsDirectory

ui_str = """<ui>
  <menubar name="MenuBar">
    <menu name="FileMenu" action="File">
      <placeholder name="FileOps_2">
        <menuitem name="QuickOpen" action="QuickOpen"/>
      </placeholder>
    </menu>
  </menubar>
</ui>
"""

class WindowHelper:
        def __init__(self, window, plugin):
                self._window = window
                self._plugin = plugin

                self._popup = None
                self._install_menu()

        def deactivate(self):
                self._uninstall_menu()
                self._window = None
                self._plugin = None

        def update_ui(self):
                pass

        def _uninstall_menu(self):
                manager = self._window.get_ui_manager()

                manager.remove_ui(self._ui_id)
                manager.remove_action_group(self._action_group)

                manager.ensure_update()

        def _install_menu(self):
                manager = self._window.get_ui_manager()
                self._action_group = Gtk.ActionGroup("PlumaQuickOpenPluginActions")
                self._action_group.add_actions([
                        ("QuickOpen", Gtk.STOCK_OPEN, _("Quick open"),
                         '<Ctrl><Alt>O', _("Quickly open documents"),
                         self.on_quick_open_activate)
                ])

                manager.insert_action_group(self._action_group, -1)
                self._ui_id = manager.add_ui_from_string(ui_str)

        def _create_popup(self):
                paths = []

                # Open documents
                paths.append(CurrentDocumentsDirectory(self._window))

                doc = self._window.get_active_document()

                # Current document directory
                if doc and doc.is_local():
                        gfile = doc.get_location()
                        paths.append(gfile.get_parent())

                # File browser root directory
                bus = self._window.get_message_bus()

                try:
                        msg = bus.send_sync('/plugins/filebrowser', 'get_root')

                        if msg:
                                uri = msg.get_value('uri')

                                if uri:
                                        gfile = Gio.file_new_for_uri(uri)

                                        if gfile.is_native():
                                                paths.append(gfile)

                except StandardError:
                        pass

                # Recent documents
                paths.append(RecentDocumentsDirectory())

                # Local bookmarks
                for path in self._local_bookmarks():
                        paths.append(path)

                # Desktop directory
                desktopdir = self._desktop_dir()

                if desktopdir:
                        paths.append(Gio.file_new_for_path(desktopdir))

                # Home directory
                paths.append(Gio.file_new_for_path(os.path.expanduser('~')))

                self._popup = Popup(self._window, paths, self.on_activated)

                self._popup.set_default_size(*self._plugin.get_popup_size())
                self._popup.set_transient_for(self._window)
                self._popup.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)

                self._window.get_group().add_window(self._popup)

                self._popup.connect('destroy', self.on_popup_destroy)

        def _local_bookmarks(self):
                filename = os.path.expanduser('~/.gtk-bookmarks')

                if not os.path.isfile(filename):
                        return []

                paths = []

                for line in file(filename, 'r').xreadlines():
                        uri = line.strip().split(" ")[0]
                        f = Gio.file_new_for_uri(uri)

                        if f.is_native():
                                try:
                                        info = f.query_info(Gio.FILE_ATTRIBUTE_STANDARD_TYPE, Gio.FileQueryInfoFlags.NONE, None)

                                        if info and info.get_file_type() == Gio.FileType.DIRECTORY:
                                                paths.append(f)
                                except GLib.GError:
                                        pass

                return paths

        def _desktop_dir(self):
                config = os.getenv('XDG_CONFIG_HOME')

                if not config:
                        config = os.path.expanduser('~/.config')

                config = os.path.join(config, 'user-dirs.dirs')
                desktopdir = None

                if os.path.isfile(config):
                        for line in file(config, 'r').xreadlines():
                                line = line.strip()

                                if line.startswith('XDG_DESKTOP_DIR'):
                                        parts = line.split('=', 1)
                                        desktopdir = os.path.expandvars(parts[1].strip('"').strip("'"))
                                        break

                if not desktopdir:
                        desktopdir = os.path.expanduser('~/Desktop')

                return desktopdir

        # Callbacks
        def on_quick_open_activate(self, action):
                if not self._popup:
                        self._create_popup()

                self._popup.show()

        def on_popup_destroy(self, popup):
                self._plugin.set_popup_size(popup.get_final_size())
                self._popup = None

        def on_activated(self, gfile):
                Pluma.commands_load_uri(self._window, gfile.get_uri(), None, -1)
                return True

# ex:ts=8:et:



#    Pluma snippets plugin
#    Copyright (C) 2005-2006  Jesse van den Kieboom <jesse@icecrew.nl>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import re

from gi.repository import GLib, Gio, Gdk, Gtk, GtkSource, Pluma

from Library import Library
from Snippet import Snippet
from Placeholder import *
import Completion

class DynamicSnippet(dict):
        def __init__(self, text):
                self['text'] = text
                self.valid = True

class Document:
        TAB_KEY_VAL = (Gdk.KEY_Tab, \
                        Gdk.KEY_ISO_Left_Tab)
        SPACE_KEY_VAL = (Gdk.KEY_space,)
        
        def __init__(self, instance, view):
                self.view = None
                self.instance = instance
                
                self.placeholders = []
                self.active_snippets = []
                self.active_placeholder = None
                self.signal_ids = {}
                
                self.ordered_placeholders = []
                self.update_placeholders = []
                self.jump_placeholders = []
                self.language_id = 0
                self.timeout_update_id = 0
                
                self.provider = Completion.Provider(_('Snippets'), self.language_id, self.on_proposal_activated)
                
                # Always have a reference to the global snippets
                Library().ref(None)
                self.set_view(view)
        
        # Stop controlling the view. Remove all active snippets, remove references
        # to the view and the plugin instance, disconnect all signal handlers
        def stop(self):
                if self.timeout_update_id != 0:
                        GLib.source_remove(self.timeout_update_id)
                        self.timeout_update_id = 0
                        del self.update_placeholders[:]
                        del self.jump_placeholders[:]

                # Always release the reference to the global snippets
                Library().unref(None)
                self.set_view(None)
                self.instance = None
                self.active_placeholder = None

        def disconnect_signal(self, obj, signal):
                if (obj, signal) in self.signal_ids:
                        obj.disconnect(self.signal_ids[(obj, signal)])
                        del self.signal_ids[(obj, signal)]
                
        def connect_signal(self, obj, signal, cb):
                self.disconnect_signal(obj, signal)     
                self.signal_ids[(obj, signal)] = obj.connect(signal, cb)

        def connect_signal_after(self, obj, signal, cb):
                self.disconnect_signal(obj, signal)
                self.signal_ids[(obj, signal)] = obj.connect_after(signal, cb)
                
        # Set the view to be controlled. Installs signal handlers and sets current
        # language. If there is already a view set this function will first remove
        # all currently active snippets and disconnect all current signals. So
        # self.set_view(None) will effectively remove all the control from the
        # current view
        def _set_view(self, view):
                if self.view:
                        buf = self.view.get_buffer()

                        # Remove signals
                        signals = {self.view: ('key-press-event', 'destroy', 
                                               'notify::editable', 'drag-data-received', 'expose-event'),
                                   buf:       ('notify::language', 'changed', 'cursor-moved', 'insert-text'),
                                   self.view.get_completion(): ('hide',)}

                        for obj, sig in signals.items():
                                if obj:
                                        for s in sig:
                                                self.disconnect_signal(obj, s)
                        
                        # Remove all active snippets
                        for snippet in list(self.active_snippets):
                                self.deactivate_snippet(snippet, True)
                        
                        completion = self.view.get_completion()
                        if completion:
                                completion.remove_provider(self.provider)

                self.view = view
                
                if view != None:
                        buf = view.get_buffer()
                       
                        self.connect_signal(view, 'destroy', self.on_view_destroy)

                        if view.get_editable():
                                self.connect_signal(view, 'key-press-event', self.on_view_key_press)

                        self.connect_signal(buf, 'notify::language', self.on_notify_language)
                        self.connect_signal(view, 'notify::editable', self.on_notify_editable)
                        self.connect_signal(view, 'drag-data-received', self.on_drag_data_received)
                        self.connect_signal_after(view, 'draw', self.on_draw)

                        self.update_language()

                        completion = view.get_completion()
                        completion.add_provider(self.provider)
                elif self.language_id != 0:
                        langid = self.language_id
                        
                        self.language_id = None;
                        self.provider.language_id = self.language_id
                        
                        if self.instance:
                                self.instance.language_changed(self)

                        Library().unref(langid)

        def set_view(self, view):
                if view == self.view:
                        return
                
                self._set_view(view)

        # Call this whenever the language in the view changes. This makes sure that
        # the correct language is used when finding snippets
        def update_language(self):
                lang = self.view.get_buffer().get_language()

                if lang == None and self.language_id == None:
                        return
                elif lang and lang.get_id() == self.language_id:
                        return

                langid = self.language_id

                if lang:
                        self.language_id = lang.get_id()
                else:
                        self.language_id = None

                if self.instance:
                        self.instance.language_changed(self)

                if langid != 0:
                        Library().unref(langid)

                Library().ref(self.language_id)
                self.provider.language_id = self.language_id

        def accelerator_activate(self, keyval, mod):
                if not self.view or not self.view.get_editable():
                        return False

                accelerator = Gtk.accelerator_name(keyval, mod)
                snippets = Library().from_accelerator(accelerator, \
                                self.language_id)

                snippets_debug('Accel!')

                if len(snippets) == 0:
                        return False
                elif len(snippets) == 1:
                        self.apply_snippet(snippets[0])
                else:
                        # Do the fancy completion dialog
                        provider = Completion.Provider(_('Snippets'), self.language_id, self.on_proposal_activated)
                        provider.set_proposals(snippets)

                        cm.show([provider], cm.create_context(None))

                return True

        def first_snippet_inserted(self):
                buf = self.view.get_buffer()
                
                self.connect_signal(buf, 'changed', self.on_buffer_changed)
                self.connect_signal(buf, 'cursor-moved', self.on_buffer_cursor_moved)
                self.connect_signal_after(buf, 'insert-text', self.on_buffer_insert_text)
        
        def last_snippet_removed(self):
                buf = self.view.get_buffer()
                self.disconnect_signal(buf, 'changed')
                self.disconnect_signal(buf, 'cursor-moved')
                self.disconnect_signal(buf, 'insert-text')

        def current_placeholder(self):
                buf = self.view.get_buffer()
                
                piter = buf.get_iter_at_mark(buf.get_insert())        
                found = []

                for placeholder in self.placeholders:
                        begin = placeholder.begin_iter()
                        end = placeholder.end_iter()

                        if piter.compare(begin) >= 0 and piter.compare(end) <= 0:
                                found.append(placeholder)

                if self.active_placeholder in found:
                        return self.active_placeholder
                elif len(found) > 0:
                        return found[0]
                else:
                        return None

        def advance_placeholder(self, direction):
                # Returns (CurrentPlaceholder, NextPlaceholder), depending on direction
                buf = self.view.get_buffer()
                
                piter = buf.get_iter_at_mark(buf.get_insert())
                found = current = next = None
                length = len(self.placeholders)
                
                placeholders = list(self.placeholders)
                
                if self.active_placeholder:
                        begin = self.active_placeholder.begin_iter()
                        end = self.active_placeholder.end_iter()
                        
                        if piter.compare(begin) >= 0 and piter.compare(end) <= 0:
                                current = self.active_placeholder
                                currentIndex = placeholders.index(self.active_placeholder)

                if direction == 1:
                        # w = piter, x = begin, y = end, z = found
                        nearest = lambda w, x, y, z: (w.compare(x) <= 0 and (not z or \
                                        x.compare(z.begin_iter()) < 0))
                        indexer = lambda x: x < length - 1
                else:
                        # w = piter, x = begin, y = end, z = prev
                        nearest = lambda w, x, y, z: (w.compare(x) >= 0 and (not z or \
                                        x.compare(z.begin_iter()) >= 0))
                        indexer = lambda x: x > 0

                for index in range(0, length):
                        placeholder = placeholders[index]
                        begin = placeholder.begin_iter()
                        end = placeholder.end_iter()
                        
                        # Find the nearest placeholder
                        if nearest(piter, begin, end, found):
                                foundIndex = index
                                found = placeholder
                        
                        # Find the current placeholder
                        if piter.compare(begin) >= 0 and \
                                        piter.compare(end) <= 0 and \
                                        current == None:
                                currentIndex = index
                                current = placeholder
                
                if current and current != found and \
                   (current.begin_iter().compare(found.begin_iter()) == 0 or \
                    current.end_iter().compare(found.begin_iter()) == 0) and \
                   self.active_placeholder and \
                   current.begin_iter().compare(self.active_placeholder.begin_iter()) == 0:
                        # if current and found are at the same place, then
                        # resolve the 'hugging' problem
                        current = self.active_placeholder
                        currentIndex = placeholders.index(current)
                
                if current:
                        if indexer(currentIndex):
                                next = placeholders[currentIndex + direction]
                elif found:
                        next = found
                elif length > 0:
                        next = self.placeholders[0]
                
                return current, next
        
        def next_placeholder(self):
                return self.advance_placeholder(1)
        
        def previous_placeholder(self):
                return self.advance_placeholder(-1)

        def cursor_on_screen(self):
                buf = self.view.get_buffer()
                self.view.scroll_mark_onscreen(buf.get_insert())
        
        def set_active_placeholder(self, placeholder):
                self.active_placeholder = placeholder

        def goto_placeholder(self, current, next):
                last = None

                if current:
                        # Signal this placeholder to end action
                        self.view.get_completion().hide()
                        current.leave()
                        
                        if current.__class__ == PlaceholderEnd:
                                last = current
                
                self.set_active_placeholder(next)
                
                if next:
                        next.enter()
                        
                        if next.__class__ == PlaceholderEnd:
                                last = next
                        elif len(next.defaults) > 1 and next.get_text() == next.default:
                                provider = Completion.Defaults(self.on_default_activated)
                                provider.set_defaults(next.defaults)
                                
                                cm = self.view.get_completion()
                                cm.show([provider], cm.create_context(None))

                if last:
                        # This is the end of the placeholder, remove the snippet etc
                        for snippet in list(self.active_snippets):
                                if snippet.placeholders[0] == last:
                                        self.deactivate_snippet(snippet)
                                        break
                
                self.cursor_on_screen()
                
                return next != None
        
        def skip_to_next_placeholder(self):
                (current, next) = self.next_placeholder()
                return self.goto_placeholder(current, next)
        
        def skip_to_previous_placeholder(self):
                (current, prev) = self.previous_placeholder()
                return self.goto_placeholder(current, prev)

        def env_get_selected_text(self, buf):
                bounds = buf.get_selection_bounds()

                if bounds:
                        return buf.get_text(bounds[0], bounds[1], False)
                else:
                        return ''

        def env_get_current_word(self, buf):
                start, end = buffer_word_boundary(buf)
                
                return buf.get_text(start, end, False)

        def env_get_current_line(self, buf):
                start, end = buffer_line_boundary(buf)
                
                return buf.get_text(start, end, False)

        def env_get_current_line_number(self, buf):
                start, end = buffer_line_boundary(buf)
                
                return str(start.get_line() + 1)
                
        def env_get_document_uri(self, buf):
                location = buf.get_location()
                
                if location:
                        return location.get_uri()
                else:
                        return ''
        
        def env_get_document_name(self, buf):
                location = buf.get_location()
                
                if location:
                        return location.get_basename()
                else:
                        return ''

        def env_get_document_scheme(self, buf):
                location = buf.get_location()
                
                if location:
                        return location.get_uri_scheme()
                else:
                        return ''

        def env_get_document_path(self, buf):
                location = buf.get_location()
                
                if location:
                        return location.get_path()
                else:
                        return ''

        def env_get_document_dir(self, buf):
                location = buf.get_location()

                if location:
                        return location.get_parent().get_path() or ''
                else:
                        return ''

        def env_get_document_type(self, buf):
                typ = buf.get_mime_type()
                
                if typ:
                        return typ
                else:
                        return ''

        def env_get_documents_uri(self, buf):
                toplevel = self.view.get_toplevel()
                
                if isinstance(toplevel, Pluma.Window):
                        documents_uri = [doc.get_location().get_uri()
                                         for doc in toplevel.get_documents()
                                         if doc.get_location() is not None]
                else:
                        documents_uri = []
                
                return ' '.join(documents_uri)

        def env_get_documents_path(self, buf):
                toplevel = self.view.get_toplevel()
                
                if isinstance(toplevel, Pluma.Window):
                        documents_location = [doc.get_location()
                                              for doc in toplevel.get_documents()
                                              if doc.get_location() is not None]

                        documents_path = [location.get_path()
                                          for location in documents_location
                                          if Pluma.utils_uri_has_file_scheme(location.get_uri())]
                else:
                        documents_path = []
                
                return ' '.join(documents_path)

        def update_environment(self):
                buf = self.view.get_buffer()
                
                variables = {'PLUMA_SELECTED_TEXT': self.env_get_selected_text, 
                             'PLUMA_CURRENT_WORD': self.env_get_current_word, 
                             'PLUMA_CURRENT_LINE': self.env_get_current_line,
                             'PLUMA_CURRENT_LINE_NUMBER': self.env_get_current_line_number,
                             'PLUMA_CURRENT_DOCUMENT_URI': self.env_get_document_uri, 
                             'PLUMA_CURRENT_DOCUMENT_NAME': self.env_get_document_name,
                             'PLUMA_CURRENT_DOCUMENT_SCHEME': self.env_get_document_scheme,
                             'PLUMA_CURRENT_DOCUMENT_PATH': self.env_get_document_path,
                             'PLUMA_CURRENT_DOCUMENT_DIR': self.env_get_document_dir,
                             'PLUMA_CURRENT_DOCUMENT_TYPE': self.env_get_document_type,
                             'PLUMA_DOCUMENTS_URI': self.env_get_documents_uri,
                             'PLUMA_DOCUMENTS_PATH': self.env_get_documents_path,
                             }
                
                for var in variables:
                        os.environ[var] = variables[var](buf)
        
        def uses_current_word(self, snippet):
                matches = re.findall('(\\\\*)\\$PLUMA_CURRENT_WORD', snippet['text'])
                
                for match in matches:
                        if len(match) % 2 == 0:
                                return True
                
                return False
        
        def uses_current_line(self, snippet):
                matches = re.findall('(\\\\*)\\$PLUMA_CURRENT_LINE', snippet['text'])
                
                for match in matches:
                        if len(match) % 2 == 0:
                                return True
                
                return False

        def apply_snippet(self, snippet, start = None, end = None):
                if not snippet.valid:
                        return False

                buf = self.view.get_buffer()
                s = Snippet(snippet)
                
                if not start:
                        start = buf.get_iter_at_mark(buf.get_insert())
                
                if not end:
                        end = buf.get_iter_at_mark(buf.get_selection_bound())

                if start.equal(end) and self.uses_current_word(s):
                        # There is no tab trigger and no selection and the snippet uses
                        # the current word. Set start and end to the word boundary so that 
                        # it will be removed
                        start, end = buffer_word_boundary(buf)
                elif start.equal(end) and self.uses_current_line(s):
                        # There is no tab trigger and no selection and the snippet uses
                        # the current line. Set start and end to the line boundary so that 
                        # it will be removed
                        start, end = buffer_line_boundary(buf)

                # Set environmental variables
                self.update_environment()
                
                # You know, we could be in an end placeholder
                (current, next) = self.next_placeholder()
                if current and current.__class__ == PlaceholderEnd:
                        self.goto_placeholder(current, None)
                
                buf.begin_user_action()

                # Remove the tag, selection or current word
                buf.delete(start, end)
                
                # Insert the snippet
                holders = len(self.placeholders)
                
                if len(self.active_snippets) == 0:
                        self.first_snippet_inserted()

                sn = s.insert_into(self, start)
                self.active_snippets.append(sn)

                # Put cursor at first tab placeholder
                keys = filter(lambda x: x > 0, sn.placeholders.keys())

                if len(keys) == 0:
                        if 0 in sn.placeholders:
                                self.goto_placeholder(self.active_placeholder, sn.placeholders[0])
                        else:
                                buf.place_cursor(sn.begin_iter())
                else:
                        self.goto_placeholder(self.active_placeholder, sn.placeholders[keys[0]])

                if sn in self.active_snippets:
                        # Check if we can get end_iter in view without moving the
                        # current cursor position out of view
                        cur = buf.get_iter_at_mark(buf.get_insert())
                        last = sn.end_iter()

                        curloc = self.view.get_iter_location(cur)
                        lastloc = self.view.get_iter_location(last)

                        if (lastloc.y + lastloc.height) - curloc.y <= \
                           self.view.get_visible_rect().height:
                                self.view.scroll_mark_onscreen(sn.end_mark)

                buf.end_user_action()
                self.view.grab_focus()

                return True

        def get_tab_tag(self, buf, end = None):
                if not end:
                        end = buf.get_iter_at_mark(buf.get_insert())

                start = end.copy()
                
                word = None
                
                if start.backward_word_start():
                        # Check if we were at a word start ourselves
                        tmp = start.copy()
                        tmp.forward_word_end()
                        
                        if tmp.equal(end):
                                word = buf.get_text(start, end, False)
                        else:
                                start = end.copy()
                else:
                        start = end.copy()
                
                if not word or word == '':
                        if start.backward_char():
                                word = start.get_char()

                                if word.isalnum() or word.isspace():
                                        return (None, None, None)
                        else:
                                return (None, None, None)

                return (word, start, end)

        def parse_and_run_snippet(self, data, iter):
                self.apply_snippet(DynamicSnippet(data), iter, iter)

        def run_snippet_trigger(self, trigger, bounds):
                if not self.view:
                        return False

                snippets = Library().from_tag(trigger, self.language_id)
                buf = self.view.get_buffer()

                if snippets:
                        if len(snippets) == 1:
                                return self.apply_snippet(snippets[0], bounds[0], bounds[1])
                        else:
                                # Do the fancy completion dialog
                                provider = Completion.Provider(_('Snippets'), self.language_id, self.on_proposal_activated)
                                provider.set_proposals(snippets)

                                cm = self.view.get_completion()
                                cm.show([provider], cm.create_context(None))

                                return True

                return False
                
        def run_snippet(self):
                if not self.view:
                        return False

                buf = self.view.get_buffer()

                # get the word preceding the current insertion position
                (word, start, end) = self.get_tab_tag(buf)

                if not word:
                        return self.skip_to_next_placeholder()

                if not self.run_snippet_trigger(word, (start, end)):
                        return self.skip_to_next_placeholder()
                else:
                        return True

        def deactivate_snippet(self, snippet, force = False):
                buf = self.view.get_buffer()
                remove = []
                ordered_remove = []

                for tabstop in snippet.placeholders:
                        if tabstop == -1:
                                placeholders = snippet.placeholders[-1]
                        else:
                                placeholders = [snippet.placeholders[tabstop]]
                        
                        for placeholder in placeholders:
                                if placeholder in self.placeholders:
                                        if placeholder in self.update_placeholders:
                                                placeholder.update_contents()
                                                
                                                self.update_placeholders.remove(placeholder)
                                        elif placeholder in self.jump_placeholders:
                                                placeholder[0].leave()

                                        remove.append(placeholder)
                                elif placeholder in self.ordered_placeholders:
                                        ordered_remove.append(placeholder)

                for placeholder in remove:
                        if placeholder == self.active_placeholder:
                                self.active_placeholder = None

                        self.placeholders.remove(placeholder)
                        self.ordered_placeholders.remove(placeholder)

                        placeholder.remove(force)

                for placeholder in ordered_remove:
                        self.ordered_placeholders.remove(placeholder)
                        placeholder.remove(force)

                snippet.deactivate()
                self.active_snippets.remove(snippet)

                if len(self.active_snippets) == 0:
                        self.last_snippet_removed()

                self.view.queue_draw()

        def update_snippet_contents(self):
                self.timeout_update_id = 0

                for placeholder in self.update_placeholders:
                        placeholder.update_contents()

                for placeholder in self.jump_placeholders:
                        self.goto_placeholder(placeholder[0], placeholder[1])

                del self.update_placeholders[:]
                del self.jump_placeholders[:]

                return False

        # Callbacks
        def on_view_destroy(self, view):
                self.stop()
                return

        def on_buffer_cursor_moved(self, buf):
                piter = buf.get_iter_at_mark(buf.get_insert())

                # Check for all snippets if the cursor is outside its scope
                for snippet in list(self.active_snippets):
                        if snippet.begin_mark.get_deleted() or snippet.end_mark.get_deleted():
                                self.deactivate(snippet)
                        else:
                                begin = snippet.begin_iter()
                                end = snippet.end_iter()
                        
                                if piter.compare(begin) < 0 or piter.compare(end) > 0:
                                        # Oh no! Remove the snippet this instant!!
                                        self.deactivate_snippet(snippet)

                current = self.current_placeholder()

                if current != self.active_placeholder:
                        self.jump_placeholders.append((self.active_placeholder, current))

                        if self.timeout_update_id == 0:
                                self.timeout_update_id = GLib.timeout_add(0, 
                                                self.update_snippet_contents)

        def on_buffer_changed(self, buf):
                for snippet in list(self.active_snippets):
                        begin = snippet.begin_iter()
                        end = snippet.end_iter()

                        if begin.compare(end) >= 0:
                                # Begin collapsed on end, just remove it
                                self.deactivate_snippet(snippet)

                current = self.current_placeholder()

                if current:
                        if not current in self.update_placeholders:
                                self.update_placeholders.append(current)
                
                        if self.timeout_update_id == 0:
                                self.timeout_update_id = GLib.timeout_add(0, \
                                                self.update_snippet_contents)
        
        def on_buffer_insert_text(self, buf, piter, text, length):
                ctx = get_buffer_context(buf)
                
                # do nothing special if there is no context and no active 
                # placeholder
                if (not ctx) and (not self.active_placeholder):
                        return
                
                if not ctx:
                        ctx = self.active_placeholder
                
                if not ctx in self.ordered_placeholders:
                        return
                        
                # move any marks that were incorrectly moved by this insertion
                # back to where they belong
                begin = ctx.begin_iter()
                end = ctx.end_iter()
                idx = self.ordered_placeholders.index(ctx)
                
                for placeholder in self.ordered_placeholders:
                        if placeholder == ctx:
                                continue
                        
                        ob = placeholder.begin_iter()
                        oe = placeholder.end_iter()
                        
                        if ob.compare(begin) == 0 and ((not oe) or oe.compare(end) == 0):
                                oidx = self.ordered_placeholders.index(placeholder)
                                
                                if oidx > idx and ob:
                                        buf.move_mark(placeholder.begin, end)
                                elif oidx < idx and oe:
                                        buf.move_mark(placeholder.end, begin)
                        elif ob.compare(begin) >= 0 and ob.compare(end) < 0 and (oe and oe.compare(end) >= 0):
                                buf.move_mark(placeholder.begin, end)
                        elif (oe and oe.compare(begin) > 0) and ob.compare(begin) <= 0:
                                buf.move_mark(placeholder.end, begin)
        
        def on_notify_language(self, buf, spec):
                self.update_language()

        def on_notify_editable(self, view, spec):
                self._set_view(view)
        
        def on_view_key_press(self, view, event):
                library = Library()

                state = event.get_state()

                if not (state & Gdk.ModifierType.CONTROL_MASK) and \
                                not (state & Gdk.ModifierType.MOD1_MASK) and \
                                event.keyval in self.TAB_KEY_VAL:
                        if not state & Gdk.ModifierType.SHIFT_MASK:
                                return self.run_snippet()
                        else:
                                return self.skip_to_previous_placeholder()
                elif not library.loaded and \
                                library.valid_accelerator(event.keyval, state):
                        library.ensure_files()
                        library.ensure(self.language_id)
                        self.accelerator_activate(event.keyval, \
                                        state & Gtk.accelerator_get_default_mod_mask())

                return False
        
        def path_split(self, path, components=[]):
                head, tail = os.path.split(path)
                
                if not tail and head:
                        return [head] + components
                elif tail:
                        return self.path_split(head, [tail] + components)
                else:
                        return components

        def relative_path(self, first, second, mime):
                prot1 = re.match('(^[a-z]+:\/\/|\/)(.*)', first)
                prot2 = re.match('(^[a-z]+:\/\/|\/)(.*)', second)

                if not prot1 or not prot2:
                        return second

                # Different protocols
                if prot1.group(1) != prot2.group(1):
                        return second

                # Split on backslash
                path1 = self.path_split(prot1.group(2))
                path2 = self.path_split(prot2.group(2))

                # Remove as long as common
                while path1 and path2 and path1[0] == path2[0]:
                        path1.pop(0)
                        path2.pop(0)

                # If we need to ../ more than 3 times, then just return
                # the absolute path
                if len(path1) - 1 > 3:
                        return second

                if mime.startswith('x-directory'):
                        # directory, special case
                        if not path2:
                                result = './'
                        else:
                                result = '../' * (len(path1) - 1)
                else:
                        # Insert ../
                        result = '../' * (len(path1) - 1)

                        if not path2:
                                result = os.path.basename(second)
                
                if path2:
                        result += os.path.join(*path2)

                return result

        def apply_uri_snippet(self, snippet, mime, uri):
                # Remove file scheme
                gfile = Gio.file_new_for_uri(uri)
                pathname = ''
                dirname = ''
                ruri = ''

                if Pluma.utils_uri_has_file_scheme(uri):
                        pathname = gfile.get_path()
                        dirname = gfile.get_parent().get_path()

                name = os.path.basename(uri)
                scheme = gfile.get_uri_scheme()

                os.environ['PLUMA_DROP_DOCUMENT_URI'] = uri
                os.environ['PLUMA_DROP_DOCUMENT_NAME'] = name
                os.environ['PLUMA_DROP_DOCUMENT_SCHEME'] = scheme
                os.environ['PLUMA_DROP_DOCUMENT_PATH'] = pathname
                os.environ['PLUMA_DROP_DOCUMENT_DIR'] = dirname
                os.environ['PLUMA_DROP_DOCUMENT_TYPE'] = mime

                buf = self.view.get_buffer()
                location = buf.get_location()
                if location:
                        ruri = location.get_uri()

                relpath = self.relative_path(ruri, uri, mime)

                os.environ['PLUMA_DROP_DOCUMENT_RELATIVE_PATH'] = relpath

                mark = buf.get_mark('gtk_drag_target')
                
                if not mark:
                        mark = buf.get_insert()

                piter = buf.get_iter_at_mark(mark)
                self.apply_snippet(snippet, piter, piter)

        def in_bounds(self, x, y):
                rect = self.view.get_visible_rect()
                rect.x, rect.y = self.view.buffer_to_window_coords(Gtk.TextWindowType.WIDGET, rect.x, rect.y)

                return not (x < rect.x or x > rect.x + rect.width or y < rect.y or y > rect.y + rect.height)
        
        def on_drag_data_received(self, view, context, x, y, data, info, timestamp):   
                uris = drop_get_uris(data)
                if not uris:
                        return

                if not self.in_bounds(x, y):
                        return

                uris.reverse()
                stop = False
                
                for uri in uris:
                        try:
                                mime = Gio.content_type_guess(uri)
                        except:
                                mime = None

                        if not mime:
                                continue
                        
                        snippets = Library().from_drop_target(mime, self.language_id)
                        
                        if snippets:
                                stop = True
                                self.apply_uri_snippet(snippets[0], mime, uri)

                if stop:
                        context.finish(True, False, timestamp)
                        view.stop_emission('drag-data-received')
                        view.get_toplevel().present()
                        view.grab_focus()
        
        def find_uri_target(self, context):
                lst = Gtk.target_list_add_uri_targets((), 0)
                
                return self.view.drag_dest_find_target(context, lst)
        
        def on_proposal_activated(self, proposal, piter):
                buf = self.view.get_buffer()
                bounds = buf.get_selection_bounds()
                
                if bounds:
                        self.apply_snippet(proposal.snippet(), None, None)
                else:
                        (word, start, end) = self.get_tab_tag(buf, piter)
                        self.apply_snippet(proposal.snippet(), start, end)

                return True
        
        def on_default_activated(self, proposal, piter):
                buf = self.view.get_buffer()
                bounds = buf.get_selection_bounds()

                if bounds:
                        buf.begin_user_action()
                        buf.delete(bounds[0], bounds[1])
                        buf.insert(bounds[0], proposal.props.label)
                        buf.end_user_action()

                        return True
                else:
                        return False

        def iter_coords(self, piter):
                rect = self.view.get_iter_location(piter)
                rect.x, rect.y = self.view.buffer_to_window_coords(Gtk.TextWindowType.TEXT, rect.x, rect.y)

                return rect

        def placeholder_in_area(self, placeholder, area):
                start = placeholder.begin_iter()
                end = placeholder.end_iter()

                if not start or not end:
                        return False

                # Test if start is before bottom, and end is after top
                start_rect = self.iter_coords(start)
                end_rect = self.iter_coords(end)

                return start_rect.y <= area.y + area.height and \
                       end_rect.y + end_rect.height >= area.y

        def draw_placeholder_rect(self, ctx, placeholder):
                start = placeholder.begin_iter()
                start_rect = self.iter_coords(start)
                start_line = start.get_line()

                end = placeholder.end_iter()
                end_rect = self.iter_coords(end)
                end_line = end.get_line()

                line = start.copy()
                line.set_line_offset(0)
                geom = self.view.get_window(Gtk.TextWindowType.TEXT).get_geometry()
                
                ctx.translate(0.5, 0.5)
                
                while line.get_line() <= end_line:
                        ypos, height = self.view.get_line_yrange(line)
                        x_, ypos = self.view.window_to_buffer_coords(Gtk.TextWindowType.TEXT, 0, ypos)

                        if line.get_line() == start_line and line.get_line() == end_line:
                                # Simply draw a box, both are on the same line
                                ctx.rectangle(start_rect.x, start_rect.y, end_rect.x - start_rect.x, start_rect.height - 1)
                                ctx.stroke()
                        elif line.get_line() == start_line or line.get_line() == end_line:
                                if line.get_line() == start_line:
                                        rect = start_rect
                                else:
                                        rect = end_rect

                                ctx.move_to(0, rect.y + rect.height - 1)
                                ctx.rel_line_to(rect.x, 0)
                                ctx.rel_line_to(0, -rect.height + 1)
                                ctx.rel_line_to(geom[2], 0)
                                ctx.stroke()

                        if not line.forward_line():
                                break

        def draw_placeholder_bar(self, ctx, placeholder):
                start = placeholder.begin_iter()
                start_rect = self.iter_coords(start)
                
                ctx.translate(0.5, 0.5)
                extend_width = 2.5

                ctx.move_to(start_rect.x - extend_width, start_rect.y)
                ctx.rel_line_to(extend_width * 2, 0)

                ctx.move_to(start_rect.x, start_rect.y)
                ctx.rel_line_to(0, start_rect.height - 1)
                
                ctx.rel_move_to(-extend_width, 0)
                ctx.rel_line_to(extend_width * 2, 0)
                ctx.stroke()

        def draw_placeholder(self, ctx, placeholder):
                if isinstance(placeholder, PlaceholderEnd):
                        return

                buf = self.view.get_buffer()

                col = self.view.get_style_context().get_color(Gtk.StateFlags.INSENSITIVE)
                col.alpha = 0.5
                Gdk.cairo_set_source_rgba(ctx, col)
                
                if placeholder.tabstop > 0:
                        ctx.set_dash([], 0)
                else:
                        ctx.set_dash([2], 0)

                start = placeholder.begin_iter()
                end = placeholder.end_iter()

                if start.equal(end):
                        self.draw_placeholder_bar(ctx, placeholder)
                else:
                        self.draw_placeholder_rect(ctx, placeholder)

        def on_draw(self, view, ctx):
                window = view.get_window(Gtk.TextWindowType.TEXT)

                if not Gtk.cairo_should_draw_window(ctx, window):
                        return False

                ctx.set_line_width(1.0)
                Gtk.cairo_transform_to_window(ctx, view, window)
                clipped, clip = Gdk.cairo_get_clip_rectangle(ctx)

                if not clipped:
                        return False

                for placeholder in self.ordered_placeholders:
                        if not self.placeholder_in_area(placeholder, clip):
                                continue

                        ctx.save()
                        self.draw_placeholder(ctx, placeholder)
                        ctx.restore()

                return False

# ex:ts=8:et:
